from django.views.generic import View
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, HttpResponseBadRequest

from .utils import is_valid_webhook
from . import signals


@method_decorator(csrf_exempt, name="dispatch")
class PaddleWebhookView(View):
    SUPPORTED_WEBHOOKS = (
        # Subscription Alerts
        "subscription_created",
        "subscription_updated",
        "subscription_cancelled",
        "subscription_payment_succeeded",
        "subscription_payment_failed",
        "subscription_payment_refunded",
        # One-off Purchases
        "locker_processed",
        "payment_succeeded",
        "payment_refunded",
        # Risk & Dispute Alerts
        "payment_dispute_created",
        "payment_dispute_closed",
        "high_risk_transaction_created",
        "high_risk_transaction_updated",
        # Payout Alerts
        "transfer_created",
        "transfer_paid",
        # Audience Alerts
        "new_audience_member",
        "update_audience_member",
    )

    def post(self, request, *args, **kwargs):
        """
        handle paddle webhook requests by
        - validating the payload signature
        - sending a django signal for each of the SUPPORTED_WEBHOOKS
        """
        payload = request.POST.dict()

        if not is_valid_webhook(payload):
            return HttpResponseBadRequest("webhook validation failed")

        alert_name = payload.get("alert_name", None)
        if alert_name is None:
            return HttpResponseBadRequest("'alert_name' missing")

        if alert_name in self.SUPPORTED_WEBHOOKS:
            signal = getattr(signals, alert_name, None)
            if signal is not None:
                signal.send(sender=self.__class__, payload=payload)

        return HttpResponse()


paddle_webhook_view = PaddleWebhookView.as_view()



from django.db import models
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _
from django.dispatch import receiver

from . import settings, signals


class Subscription(models.Model):
    """
    'Subscription' represents a Paddle subscription.

    They are automatically updated on behalf of the webhook calls 'subscription_created',
    'subscription_updated' and 'subscription_cancelled'.

    Stale subscriptions that are not associated with a subscriber are being linked by comparing
    the email addresses of the subscriber model and the subscription. Linking is optional and can
    be configured through the settings ('DJPADDLE_LINK_STALE_SUBSCRIPTIONS').
    """

    STATUS_ACTIVE = "active"
    STATUS_TRIALING = "trialing"
    STATUS_PAST_DUE = "past_due"
    STATUS_PAUSED = "paused"
    STATUS_DELETED = "deleted"
    STATUS = (
        (STATUS_ACTIVE, _("active")),
        (STATUS_TRIALING, _("trialing")),
        (STATUS_PAST_DUE, _("past due")),
        (STATUS_PAUSED, _("paused")),
        (STATUS_DELETED, _("deleted")),
    )

    id = models.CharField(max_length=32, primary_key=True)
    subscriber = models.ForeignKey(
        settings.DJPADDLE_SUBSCRIBER_MODEL,
        related_name="subscriptions",
        null=True,
        default=None,
        on_delete=models.CASCADE,
    )

    cancel_url = models.URLField()
    checkout_id = models.CharField(max_length=32)
    currency = models.CharField(max_length=3)
    email = models.EmailField()
    event_time = models.DateTimeField()
    marketing_consent = models.BooleanField()
    next_bill_date = models.DateTimeField()
    passthrough = models.TextField()
    quantity = models.IntegerField()
    source = models.URLField()
    status = models.CharField(choices=STATUS, max_length=16)
    subscription_plan_id = models.CharField(max_length=32)
    unit_price = models.FloatField()
    update_url = models.URLField()

    @classmethod
    def _sanitize_webhook_payload(cls, payload):
        data = {}

        data["id"] = data.pop("subscription_id", None)

        # transform `user_id` to subscriber ref
        data["subscriber"] = None
        subscriber_id = data.pop("user_id", None)
        if subscriber_id not in ["", None]:
            data["subscriber"] = settings.get_subscriber_model().objects.get(
                pk=subscriber_id
            )

        # sanitize fields
        valid_field_names = [field.name for field in cls._meta.get_fields()]
        for key, value in payload.items():
            # e.g. 'new_status' --> 'status'
            if key.startswith("new_"):
                key = key[4:]

            if key in valid_field_names:
                data[key] = value

        return data

    @classmethod
    def from_subscription_created(cls, payload):
        data = cls._sanitize_webhook_payload(payload)
        return cls(**data)

    @classmethod
    def update_by_payload(cls, payload):
        data = cls._sanitize_webhook_payload(payload)
        pk = data.pop("id")
        return cls.objects.update_or_create(pk, defaults=data)


@receiver(signals.subscription_created)
def on_subscription_created(sender, payload, *args, **kwargs):
    Subscription.from_subscription_created(payload)


@receiver(signals.subscription_updated)
def on_subscription_updated(sender, payload, *args, **kwargs):
    Subscription.update_by_payload(payload)


@receiver(signals.subscription_cancelled)
def on_subscription_cancelled(sender, payload, *args, **kwargs):
    Subscription.update_by_payload(payload)


if settings.DJPADDLE_LINK_STALE_SUBSCRIPTIONS:

    @receiver(post_save, sender=settings.DJPADDLE_SUBSCRIBER_MODEL)
    def link_stale_subscriptions_to_subscriber(
        sender, instance, created, *args, **kwargs
    ):
        if created:
            subscriptions = Subscription.objects.filter(
                user=None, email__iexact=instance.email
            )
            subscriptions.update(user=instance)



import base64
import hashlib
import collections
import phpserialize

from Crypto.PublicKey import RSA
from django.core.exceptions import ImproperlyConfigured

try:
    from Crypto.Hash import SHA1
except ImportError:
    from Crypto.Hash import SHA as SHA1

try:
    from Crypto.Signature import PKCS1_v1_5
except ImportError:
    from Crypto.Signature import pkcs1_15 as PKCS1_v1_5

from . import settings


def convert_pubkey_to_rsa(key):
    """
    convert key from PEM to DER - Strip the first and last lines and newlines, decode and
    return an instance of `Crypto.PublicKey.RSA`
    """
    public_key_encoded = "".join(key.split("\n")[1:-1])
    public_key_der = base64.b64decode(public_key_encoded)
    return RSA.importKey(public_key_der)


try:
    key = convert_pubkey_to_rsa(settings.DJPADDLE_PUBLIC_KEY)
except Exception as e:
    msg = "failed to convert 'DJPADDLE_PUBLIC_KEY'; original message: " + str(e)
    raise ImproperlyConfigured(msg)

verifier = PKCS1_v1_5.new(key)


def is_valid_webhook(payload):
    data = dict(payload)
    signature = data.pop("p_signature", None)
    if signature is None:
        return False

    # Ensure all the data fields are strings
    for field in data:
        data[field] = str(data[field])

    # Sort the data and serialize it
    sorted_data = collections.OrderedDict(sorted(data.items()))
    serialized_data = phpserialize.dumps(sorted_data)

    # verify the data
    digest = SHA1.new()
    digest.update(serialized_data)
    signature = base64.b64decode(signature)

    return verifier.verify(digest, signature)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.apps import apps as django_apps


# can be found at https://vendors.paddle.com/authentication
DJPADDLE_VENDOR_ID = getattr(settings, "DJPADDLE_VENDOR_ID", None)

# create one at https://vendors.paddle.com/authentication
DJPADDLE_API_KEY = getattr(settings, "DJPADDLE_API_KEY", None)

# can be found at https://vendors.paddle.com/public-key
DJPADDLE_PUBLIC_KEY = getattr(settings, "DJPADDLE_PUBLIC_KEY", None)
if DJPADDLE_PUBLIC_KEY is None:
    raise ImproperlyConfigured("'DJPADDLE_PUBLIC_KEY' must be set")

DJPADDLE_SUBSCRIBER_MODEL = getattr(
    settings, "DJPADDLE_SUBSCRIBER_MODEL", settings.AUTH_USER_MODEL
)

DJPADDLE_LINK_STALE_SUBSCRIPTIONS = getattr(
    settings, "DJPADDLE_LINK_STALE_SUBSCRIPTIONS", True
)


def get_subscriber_model():
    """
    Attempt to pull settings.DJPADDLE_SUBSCRIBER_MODEL.
    Users have the option of specifying a custom subscriber model via the
    DJPADDLE_SUBSCRIBER_MODEL setting.
    This methods falls back to AUTH_USER_MODEL if DJPADDLE_SUBSCRIBER_MODEL is not set.
    Returns the subscriber model that is active in this project.
    """
    model_name = DJPADDLE_SUBSCRIBER_MODEL

    # Attempt a Django 1.7 app lookup
    try:
        subscriber_model = django_apps.get_model(model_name)
    except ValueError:
        raise ImproperlyConfigured(
            "DJPADDLE_SUBSCRIBER_MODEL must be of the form 'app_label.model_name'."
        )
    except LookupError:
        raise ImproperlyConfigured(
            "DJPADDLE_SUBSCRIBER_MODEL refers to model '{model}' "
            "that has not been installed.".format(model=model_name)
        )

    if (
        "email" not in [field_.name for field_ in subscriber_model._meta.get_fields()]
    ) and not hasattr(subscriber_model, "email"):
        raise ImproperlyConfigured(
            "DJPADDLE_SUBSCRIBER_MODEL must have an email attribute."
        )

    return subscriber_model



import django.dispatch

# Subscription Alerts
subscription_created = django.dispatch.Signal(providing_args=["payload"])
subscription_updated = django.dispatch.Signal(providing_args=["payload"])
subscription_cancelled = django.dispatch.Signal(providing_args=["payload"])
subscription_payment_succeeded = django.dispatch.Signal(providing_args=["payload"])
subscription_payment_failed = django.dispatch.Signal(providing_args=["payload"])
subscription_payment_refunded = django.dispatch.Signal(providing_args=["payload"])

# One-off Purchases
locker_processed = django.dispatch.Signal(providing_args=["payload"])
payment_succeeded = django.dispatch.Signal(providing_args=["payload"])
payment_refunded = django.dispatch.Signal(providing_args=["payload"])

# Risk & Dispute Alerts
payment_dispute_created = django.dispatch.Signal(providing_args=["payload"])
payment_dispute_closed = django.dispatch.Signal(providing_args=["payload"])
high_risk_transaction_created = django.dispatch.Signal(providing_args=["payload"])
high_risk_transaction_updated = django.dispatch.Signal(providing_args=["payload"])

# Payout Alerts
transfer_created = django.dispatch.Signal(providing_args=["payload"])
transfer_paid = django.dispatch.Signal(providing_args=["payload"])

# Audience Alerts
new_audience_member = django.dispatch.Signal(providing_args=["payload"])
update_audience_member = django.dispatch.Signal(providing_args=["payload"])



#!/usr/bin/env python
import os
import sys

from setuptools import setup

if sys.argv[-1] == "publish":
    os.system("python setup.py bdist_wheel upload --sign")
    sys.exit()

README = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()

setup(long_description=README)
